

#    This file is common to both AGATADigitizer.tml and Diagnostics.tml
#    However it is loaded into the appropiate namespace


proc EnumerateDigitizers {} {
    variable Digitizer
    variable DigitizerList
    global errorCode errorInfo

    variable Detector
    variable DetectorList

    variable Crystal
    variable CrystalList

    variable Segment
    variable SegmentList

    variable Segment2Fpga
    variable Fpga2Segment

    variable JS

    upvar #0 AGATADigitizerAccessServer_LayoutFile LayoutFile
    upvar #0 AGATADigitizerAccessServer_Detector2Digitizer Detector2Digitizer
    upvar #0 AGATADigitizerAccessServer_Digitizer2Detector Digitizer2Detector

    set DigitizerList  {}
    set NumDigitizer 0
    set Digitizer -1

    set DetectorList {}
    set CrystalList {}
    set SegmentList {}

    set Detector ""
    set Crystal ""
    set Segment ""

    set rc [catch {::namespace inscope ::urn:AGATADigitizerAccessServer do_Init} m]
    if {$rc != 0} {
        append JS "alert(\"Init returned with an error\");"
        InsertLog "Init returned with an error\n$m\n$errorCode\n$errorInfo"
        return
    } else {
        set DigitizerList $m
        set NumDigitizer [llength $DigitizerList]
        set Digitizer [first $DigitizerList]
        InsertLog "discovered $NumDigitizer digitizer(s)"
    }

    set z [catch {source $LayoutFile} m]
    if {$z != 0} {
        append JS "alert(\"Unable to access $LayoutFile\");"
        InsertLog "[set m]"
        return
    }

    foreach item [array names Detector2Digitizer] {
        set Digitizer2Detector([set Detector2Digitizer($item)]) $item

    }

    foreach item [array names Segment2Fpga] {
        set Fpga2Segment([set Segment2Fpga($item)]) $item

    }

    set item "Core: Main Board"
    set Fpga2Segment($item) ""

    set item "Segment: Main Board"
    set Fpga2Segment($item) ""


    foreach item [array names Detector2Digitizer] {
         set detector [first [split $item _]]
         set crystal [second [split $item _]]
         if {[lsearch $DetectorList $detector] == -1} {lappend DetectorList $detector}
         if {[lsearch $CrystalList $crystal] == -1} {lappend CrystalList $crystal}
    }


    foreach item [array names Segment2Fpga] {
         lappend SegmentList $item
    }
    
    set Detector [first [split [set Digitizer2Detector($Digitizer)] _]]
    set Crystal [second [split [set Digitizer2Detector($Digitizer)] _]]
    set Segment I

}

proc click-probe {} {
    variable Digitizer 
    variable Present 
    variable CorePresent
    variable SegmentPresent 
    variable FpgaList
    variable FpgaIDs 
    variable Fpga 
    variable ShortTimeOut 

    set CorePresent 0
    set SegmentPresent 0
    for {set i 0} {$i <= 12} {incr i} {set Present($i) 0} 

    InsertLog "Digitizer= $Digitizer"

    set FpgaList {}
    set FpgaIDs {}

    set fpga 3     ;# core main board
    set cmd 14     ;# safe read command for test
    set Length 6   ;# returns 6 bytes

    set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]

    for {set tries 1} {$tries <= 2} {incr tries} {

    InsertLog "probing for Core Main Board"

    set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $cmd A16 $Length} m]
    if {$rc == 0} {
        InsertLog "found it"
        lappend FpgaList "Core: Main Board"
        lappend FpgaIDs $fpga
        set CorePresent 1
        set Present($fpga) 1
        break
    } else {
        InsertLog "Error returned [first [split $m \n]]"
    }
    }

    if {$CorePresent == 1} {

       InsertLog "probing for Core ADCs"

       set fpga 2    ;# core ADCs
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Core: Virtex for core ADCs"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

       InsertLog "probing for Core segment ADC card #1"

       set fpga 0    ;# core segment ADC card #1
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Core: Virtex for Segment ADC card 1"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

       InsertLog "probing for Core segment ADC card #2"

       set fpga 1    ;# core segment ADC card #2
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Core: Virtex for Segment ADC card 2"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }
    }

    InsertLog "probing for Segment Main Board"

    set fpga 12    ;# segment main board
    set cmd 14     ;# safe read command for test
    set Length 6   ;# returns 6 bytes
    set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
    set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $cmd A16 $Length} m]
    if {$rc == 0} {
        InsertLog "found it"
        lappend FpgaList "Segment: Main Board"
        lappend FpgaIDs $fpga
        set SegmentPresent 1
        set Present($fpga) 1
    } else {
        InsertLog "Error returned [first [split $m \n]]"
    }

    if {$SegmentPresent == 1} {

       InsertLog "probing for Segment ADC card #1"

       set fpga 8    ;# Segment ADC card #1
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 1"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

       InsertLog "probing for Segment ADC card #2"

       set fpga 9    ;# Segment ADC card #2
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 2"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

       InsertLog "probing for Segment ADC card #3"

       set fpga 10    ;# Segment ADC card #3
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 3"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

       InsertLog "probing for Segment ADC card #4"

       set fpga 11    ;# Segment ADC card #4
       set address 15
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           InsertLog "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 4"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           InsertLog "Error returned [first [split $m \n]]"
       }

   }

   set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_SetTimeOut $Digitizer 0} m]   ;# reset timeout to default

   if {[expr $CorePresent + $SegmentPresent] == 0} {
       InsertLog "Digitizer not found"
       set FpgaList {"Digitizer not found"}
       set Digitizer -1
   } else {
       InsertLog "Found Digitizer $Digitizer which consists of:"
       foreach item $FpgaList {InsertLog $item}
   }
   set Fpga [first $FpgaList]

}

proc click-readprom {} {
        variable JS
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs
        variable  Address
        variable  AddressModifier 
        global errorCode errorInfo

set Prom_control 0x880
set Prom_status 0x881
set Prom_base 0x800

        if {$Digitizer == -1} {
            InsertLog "Use \"Probe Hardware\" and select a Digitizer"
            set JS "alert(\"Use Probe Hardware and select a Digitizer\");"
            return
        }

        if {$Fpga == "Use Probe Hardware"} {
            InsertLog "Use \"Probe Hardware\" and select a FPGA"
            set JS "alert(\"Use Probe Hardware and select a FPGA\");"
            return
        }

    InsertLog "Digitizer= $Digitizer"

    InsertLog "Read PROM"

#    not available for Main Boards

        if {$Fpga == "Core: Main Board" || $Fpga == "Segment: Main Board"} {
            InsertLog "Not available for Main Boards"
            set JS "alert(\"Not available for Main Boards\");"
            return
        }

# check to see if it has read in correctly 

        set a [ Get_Register $Prom_status ]
	if { $a != 4 } {
             InsertLog "Prom has read back unsuccessfully. Status : $a"
             return
        }

	InsertLog "Prom Read back into RAM"

	set a [ format "%c" [ Get_Register 0x800 ] ]
	
	if { $a == "C" } {
		InsertLog "This is a Core ADC board"
	} elseif { $a == "S" } {
		InsertLog "This is a Segment ADC board" 
	} else {
		InsertLog "Not sure what this is : $a [ Get_Register 0x800 ] !"
                set JS "alert(\"Not sure what this is. See log for further information\");"
                return
	}
	set a [ Get_Register 0x801 ]
	set b [ Get_Register 0x802 ]
	set c [ Get_Register 0x803 ]
	set d [ Get_Register 0x804 ]
	set e [ Get_Register 0x805 ]
	set f [ Get_Register 0x806 ]
	set g [ Get_Register 0x807 ]
	set serial [ format "%c%c%c%c" $d $c $b $a ]
	set revision [ format "%c" $e ]
	set mod_state [ format "%c%c" $g $f ]

	InsertLog "Serial number : $serial"
	InsertLog "Revision      : $revision"
	InsertLog "Mod state     : $mod_state"
	InsertLog "Test Complete ....."

}

proc do_UpdateOffsets {} {
        variable  Fpga 
        variable Offset

        if {$Fpga == "Core: Main Board" || $Fpga == "Segment: Main Board"} {return}   ;# Main boards have no offset registers

        if {$Fpga == "Core: Virtex for core ADCs"} {
            for {set i 1} {$i <= 2} {incr i} {set Offset($i) [Get_Register [expr 0x80 + $i]]}
        } else {
            for {set i 1} {$i <= 7} {incr i} {set Offset($i) [Get_Register [expr 0x80 + $i]]}
        }

     for {set i 1} {$i <= 7} {incr i} {if {[set Offset($i)] > 32767} {set Offset($i) [expr [set Offset($i)] - 65536]}}  ;# fixup till we understand signed offsets
}

proc click-offset {vl} {

#    vl = 0    write current FPGA offsets to the hardware
#       = 1    copy from database to hardware
#       = 2    copy from hardware to database

    if {$vl == 0} {do-SetOffsets; return}
    if {$vl == 1} {do-RestoreOffsets; return}
    if {$vl == 2} {do-SaveOffsets; return}
}

proc do-SetOffsets {} {
        variable Fpga 
        variable Offset

        if {$Fpga == "Core: Main Board" || $Fpga == "Segment: Main Board"} {
            InsertLog "Not available for Main Boards"
            return
        }

        if {$Fpga == "Core: Virtex for core ADCs"} {
            for {set i 1} {$i <= 2} {incr i} {
                Set_Register [expr 0x80 + $i] [set Offset($i)]
                InsertLog "Set Offset $i = [set Offset($i)]"
            }
        } else {
            for {set i 1} {$i <= 7} {incr i} {
                Set_Register [expr 0x80 + $i] [set Offset($i)]
                InsertLog "Set Offset $i = [set Offset($i)]"
            }
        }
}

proc do-RestoreOffsets {} {
        upvar #0 AGATADigitizerAccessServer_ConfigFile ConfigFile
        variable  Digitizer
        variable  Fpga
        variable JS

        set FpgaBAK $Fpga

        InsertLog "Restoring offsets for digitizer [set Digitizer]"

        set SaveFile [file join [file dirname $ConfigFile] AGATADigitizerOffsets[set Digitizer].dat]

        set z [catch {open $SaveFile r} file]
        if {$z != 0} {
            append JS "alert(\"Unable to access $SaveFile\");"
            InsertLog "[set file]"
            return
        }

        foreach item [split [read $file] \n] {
            if {[string length $item] == 0} {continue}
            set Fpga [first [split $item &]]
            set i [second [split $item &]]
            set vl  [last [split $item &]]
            Set_Register [expr 0x80 + $i] $vl
            InsertLog "fpga $Fpga offset $i value $vl"
        }

        close $file

        InsertLog "Done"

        Elog INFORMATION "Restored offsets for digitizer [set Digitizer] from $SaveFile"

        set Fpga $FpgaBAK
        do_UpdateOffsets
}

proc do-SaveOffsets {} {
        upvar #0 AGATADigitizerAccessServer_ConfigFile ConfigFile
        variable  Digitizer
        variable  Fpga
        variable  FpgaList
        variable  Offset
        variable JS

        set FpgaBAK $Fpga

        InsertLog "Saving offsets for digitizer [set Digitizer]"

        set SaveFile [file join [file dirname $ConfigFile] AGATADigitizerOffsets[set Digitizer].dat]

        set z [catch {open $SaveFile w} file]
        if {$z != 0} {
            append JS "alert(\"Unable to access $SaveFile\");"
            InsertLog "[set file]"
            return
        }

        foreach item $FpgaList {
            set Fpga $item
            if {$Fpga == "Core: Main Board"} {continue}   ;# Main boards have no offset registers
            if {$Fpga == "Segment: Main Board"} {continue}   ;# Main boards have no offset registers
            do_UpdateOffsets 
            if {$Fpga == "Core: Virtex for core ADCs"} {
                for {set i 1} {$i <= 2} {incr i} {puts $file "$Fpga&$i&[set Offset($i)]"; InsertLog "fpga $Fpga offset $i value [set Offset($i)]"}
            } else {
                for {set i 1} {$i <= 7} {incr i} {puts $file "$Fpga&$i&[set Offset($i)]"; InsertLog "fpga $Fpga offset $i value [set Offset($i)]"}
            }
        }

        close $file

        InsertLog "Done"

        Elog INFORMATION "Offsets for digitizer [set Digitizer] updated in $SaveFile"

        set Fpga $FpgaBAK
        do_UpdateOffsets
}

proc do_SetUpAllDigitizers {} {click-setupall}      ;#  nice name for Acquisition control

proc click-setupall {} {
    variable Digitizer
    variable DigitizerList

    set DigitizerBAK $Digitizer

    EmptyLog

    foreach Digitizer $DigitizerList  {
        click-probe
        click-setupdigitizer
    }

    PrintLog

    EmptyLog
    InsertLog "Setup all complete"

    set Digitizer $DigitizerBAK
}

proc click-checkall {} {
    variable Digitizer
    variable DigitizerList
    upvar #0 AGATADigitizerAccessServer_ConfigFile ConfigFile

    set DigitizerBAK $Digitizer

    set CheckFile AGATADigitizer.chk

    EmptyLog

    InsertLog "Checking digitizers using $CheckFile"

    foreach Digitizer $DigitizerList  {
        click-probe
        click-checkdigitizer
    }

    InsertLog "Checks all complete"
    PrintLog

    set Digitizer $DigitizerBAK
}

proc click-setupdigitizer {} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address 
        variable  AddressModifier

        set FpgaBAK $Fpga

# set the external clock option
	if { [SelectFpga "Core: Main Board"] == -1 } then { 
		InsertLog "Can't connect to Main board" 
		return
	}
	set rtnd 0
	Set_Register 40 0 

# enable the Core Laser and push it's DCMs to lock.

	if { [SelectFpga "Core: Virtex for core ADCs"] == -1 } then { 
		InsertLog "No Core card available !" 
		return
	}

# set the sync pulse to be correctly transmitted

	Set_Register 11 1
	
# set the basics - inspection and main enables
	Set_Register 4 0

	Set_Register 7 0x300
	Set_Register 0 0x175
	Set_Register 0 0x75
 	set val [Get_Register 1 ]
	set val [ expr $val & 0xfffD ]
 	if { $val != 0xf205 } then {
		InsertLog "Core status  fault : ([format "0x%x" $val])"
		InsertLog "Is the cable connected ?"
	}
	setup_core  
	setup_segment "Core: Virtex for Segment ADC card 1" 
	setup_segment "Core: Virtex for Segment ADC card 2" 
	setup_segment "Segment: Virtex for Segment ADC card 1" 
	setup_segment "Segment: Virtex for Segment ADC card 2" 
	setup_segment "Segment: Virtex for Segment ADC card 3" 
	setup_segment "Segment: Virtex for Segment ADC card 4" 

        set Fpga $FpgaBAK

    do-RestoreOffsets

      InsertLog "Setup Complete ....."
}

proc click-checkdigitizer {} {
        upvar #0 AGATADigitizerAccessServer_ConfigFile ConfigFile
        variable  Digitizer
        variable  Fpga
        variable FpgaList 
        variable FpgaIDs 
        global errorCode errorInfo
        variable JS

        set FpgaBAK $Fpga

        set CheckFile AGATADigitizer.chk

        set z [catch {open $CheckFile r} file]
        if {$z != 0} {
            append JS "alert(\"Unable to access $CheckFile\");"
            InsertLog "[set file]"
            return
        }

        InsertLog "Performing checks for digitizer [set Digitizer]"

        foreach item [split [read $file] \n] {
            if {[string length $item] == 0} {continue}
            if {[string range $item 0 0] == "#"} {continue}
            if {[SelectFpga [first [split $item &]]] == -1} then {continue}
            set Offset [second [split $item &]]
            set Mask [third [split $item &]]
            set Value [fourth [split $item &]]
            set Text  [fifth [split $item &]]
            if {[string compare -length 16 $Fpga "Core: Main Board"] == 0 || [string compare -length 19 $Fpga "Segment: Main Board"] == 0} {
                set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]
                set cmd 14     ;# read status registers command
                set Length 6   ;# returns 6 bytes
                set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $cmd A16 $Length} m]
                if {$rc == 0} {
                    set vl [lindex $m $Offset]
                } else {
                    InsertLog "do_Read returned with an error\n$m\n$errorCode\n$errorInfo"
                    set vl -1
                }
            } else {
                set vl [Get_Register $Offset]
            }
            if {[expr $vl & $Mask] != $Value} {
                 InsertLog "$Fpga: $Text: Obtained [format 0x%x [expr $vl & $Mask]]; expected $Value"
            }
#            InsertLog "fpga= $Fpga; offset= $Offset; mask= $Mask; value= $Value; obtained [format 0x%x $vl]"
        }

        close $file

        InsertLog "Done"

        set Fpga $FpgaBAK
}

proc click-setupfpga {} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs
        variable  Address
        variable  AddressModifier 

        if {$Fpga == "Core: Main Board" || $Fpga == "Segment: Main Board"} {
            InsertLog "Not available for Main Boards"
            return
        }

        if {$Fpga == "Core: Virtex for core ADCs"} {setup_core} else {setup_segment $Fpga}

        InsertLog "Setup $Fpga Complete ....."

}

proc setup_segment { board } {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address
        variable  AddressModifier 

	set rtnd 0

	if { [SelectFpga "$board"] == -1 } then { 
		InsertLog "Segment card --- %board---- not available !" 
		return
	}
# set the basics  disable inspection and main enables
# enable the laser
	Set_Register 0 5
	Set_Register 4 0
	Set_Register 7 0xFF00

# check the status register for DCM locked.
	set val [Get_Register 1 ]
	if { $val != 0xe001 } {
		InsertLog "Segment status  fault : ([format "0x%x" $val])"
	        return
	}
# channel by channel set transmit data enabled
	foreach val { 1 2 3 4 5 6 } {
		set ch_rkt [expr ( $val << 4 ) + 3 ]
		Set_Register $ch_rkt 0x7c00
		set ch_rkt [expr ( $val << 4 ) + 4 ]
		Set_Register $ch_rkt 0
	}
# aving these as two separate or statements allows link synch.
	foreach val { 1 2 3 4 5 6 } {
		set ch_rkt [expr ( $val << 4 ) + 3 ]
		Set_Register $ch_rkt 0
	}
	
}

proc setup_core {} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address
        variable  AddressModifier 

	if { [SelectFpga "Core: Virtex for core ADCs"] == -1 } then { 
		WaitForUser "No Core card available !" 
		return
	}
# set the basics - inspection and main enables
	Set_Register 4 0
	Set_Register 7 0x300

# channel by channel set transmit data enabled
	foreach val { 1 2  } {
		set ch_rkt [expr ( $val << 4 ) + 3 ]
		Set_Register $ch_rkt 0x7c00
		set ch_rkt [expr ( $val << 4 ) + 4 ]
		Set_Register $ch_rkt 0
	}
	foreach val { 1 2  } {
		set ch_rkt [expr ( $val << 4 ) + 3 ]
		Set_Register $ch_rkt 0
	}
}


    proc SelectFpga {n} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address
        variable  AddressModifier 

       if {[lsearch $FpgaList $n] == -1 || [lindex $FpgaIDs [lsearch $FpgaList $n]] == -1} {
           InsertLog "$n is a bad Fpga"
           return -1
       } else {
           set Fpga $n
       }
       return [set Fpga]
    }

    proc Get_Register {Addr {Space A16}} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs
        variable  Address 
        variable  AddressModifier 
        global errorCode errorInfo

        set Address $Addr
        set AddressModifier $Space

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        if {$AddressModifier == "A16"} {
            set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $Address $AddressModifier} m]
        } else {
            set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $Address $AddressModifier 1} m]
        }
        if {$rc != 0} {
            InsertLog "Get_Register: $Digitizer $fpga $Address $AddressModifier"
            InsertLog "Get_Register returned with an error\n$m\n$errorCode\n$errorInfo"
            return -1
        } else {
            return $m
        }
   }

   proc Set_Register {Addr Data {Space A16}}  {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address 
        variable  AddressModifier 
        global errorCode errorInfo

        set Address $Addr
        set AddressModifier $Space

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Set $Digitizer $fpga $Address $AddressModifier 1 $Data} m]
        if {$rc != 0} {
            InsertLog "Set_Register: $Digitizer $fpga $Address $AddressModifier 1 $Data"
            InsertLog "Set_Register returned with an error\n$m\n$errorCode\n$errorInfo"
            return -1
        } else {
            return $m
        }
   }

    proc do_get {} {
        variable  Digitizer 
        variable  Fpga 
        variable  FpgaList 
        variable  FpgaIDs 
        variable  Address 
        variable  AddressModifier 
        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        if {$AddressModifier == "A16"} {
            set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Get $Digitizer $fpga $Address A16} m]
        } else {
            set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $Address A32 1} m]
        }
        if {$rc != 0} {
            InsertLog "Get returned with an error\n$m\n$errorCode\n$errorInfo"
        } else {
            InsertLog "Get returned $m ([format "0x%x" $m])"
        }
    }

    proc do_read {} {
        variable Digitizer 
        variable Fpga 
        variable FpgaList 
        variable FpgaIDs 
        variable Address 
        variable Length 
        variable AddressModifier
        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $Address $AddressModifier $Length} m]
        if {$rc != 0} {
            InsertLog "Read returned with an error\n$m\n$errorCode\n$errorInfo"
        } else {
            InsertLog "Read returned $m"
        }
    }

    proc do_write {} {
        variable Digitizer
        variable Fpga 
        variable FpgaList 
        variable FpgaIDs
        variable Address 
        variable DataBlock 
        variable AddressModifier 
        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Set $Digitizer $fpga $Address $AddressModifier [llength $DataBlock] $DataBlock} m]
        if {$rc != 0} {
            InsertLog "Write returned with an error\n$m\n$errorCode\n$errorInfo"
        } else {
            InsertLog "Write returned $m ([format "0x%x" $m])"
        }
    }

    proc do_shutdown {} {
        variable Digitizer 
        variable Fpga 
        variable FpgaList 
        variable FpgaIDs
        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]
        if {$fpga < 8} then {set fpga 3} else {set fpga 12}    ;# select main board

        set cmd 20     ;# shutdown command

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Set $Digitizer $fpga $cmd A16 1 0x0800} m]
        if {$rc == 0} {
            InsertLog "Done"
        } else {
            InsertLog "Write returned with an error\n$m\n$errorCode\n$errorInfo"
        }
    }

    proc do_showstatus {} {
        variable Digitizer 
        variable Fpga
        variable FpgaList 
        variable FpgaIDs 
        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]
        if {$fpga < 8} then {set fpga 3} else {set fpga 12}    ;# select main board

        set cmd 14     ;# read status registers command
        set Length 6   ;# returns 6 bytes
        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Read $Digitizer $fpga $cmd A16 $Length} m]
        if {$rc == 0} {
            InsertLog "status registers: $m"
        } else {
            InsertLog "Write returned with an error\n$m\n$errorCode\n$errorInfo"
        }
    }

    proc do_clock {vl} {
        variable Digitizer
        global errorCode errorInfo

        set fpga 3    ;#  available to core main board only
        set cmd 40

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Set $Digitizer $fpga $cmd A16 1 $vl} m]
        if {$rc != 0} {
            InsertLog "Set returned with an error\n$m\n$errorCode\n$errorInfo"
        } else {
            InsertLog "Set returned $m ([format "0x%x" $m])"
        }
    }

    proc do_set {} {
        variable Digitizer
        variable Fpga
        variable Address
        variable AddressModifier
        variable Data
        variable Length
        variable DataBlock

        variable FpgaList
        variable FpgaIDs

        global errorCode errorInfo

        set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

        set rc [catch {interp eval "" ::namespace inscope ::urn:AGATADigitizerAccessServer do_Set $Digitizer $fpga $Address $AddressModifier 1 $Data} m]
        if {$rc != 0} {
            InsertLog "Set returned with an error\n$m\n$errorCode\n$errorInfo"
        } else {
            InsertLog "Set returned $m ([format "0x%x" $m])"
        }
    }

