
#    MIDAS Access Server  for  AGATA Digitizer 

global env
set z [catch {source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]} m]       ;#  load default support procedures
if {$z != 0} {puts "software load failure: $m"}

set z [catch {source [file join $env(MIDASBASE) SOAP tcl MessageLoggerClient.tcl]} m] ;#  load Message Logger Client
if {$z != 0} {puts "software load failure: $m"}

set z [catch {source [file join $env(MIDASBASE) MIDASpackages XportAccess1.0 pkgIndex.tcl]} m] ;#  load Xport access library
if {$z != 0} {puts "plugin load failure: $m"}



global MIDAS_AGATADigitizerAccessServer_Verbose
set MIDAS_AGATADigitizerAccessServer_Verbose 0

global MIDAS_SOAPDomainHandler_Verbose
set MIDAS_SOAPDomainHandler_Verbose 0

global MIDAS_RegisterAccessServer_Verbose
set MIDAS_RegisterAccessServer_Verbose 0

upvar #0 AGATADigitizerAccessServer_Connected Connected
upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut

upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
set MaxDigitizer 16

upvar #0 AGATADigitizerAccessServer_NumDigitizer NumDigitizer
set NumDigitizer 0


    upvar #0 AGATADigitizerAccessServer_ConfigFile ConfigFile
    upvar #0 AGATADigitizerAccessServer_LayoutFile LayoutFile

    upvar #0 AGATADigitizerAccessServer_XportList XportList
    upvar #0 AGATADigitizerAccessServer_InUse InUse
    upvar #0 AGATADigitizerAccessServer_Detector2Digitizer Detector2Digitizer
    upvar #0 AGATADigitizerAccessServer_Digitizer2Detector Digitizer2Detector
    upvar #0 AGATADigitizerAccessServer_Segment2Fpga Segment2Fpga

    source $ConfigFile
    source $LayoutFile


for {set i 0} {$i < $MaxDigitizer} {incr i} {set InUse($i) 0; set Connected($i) 0; set TimeOut($i) 0}

for {set i 0} {$i < $MaxDigitizer} {incr i} {
    if {![info exists XportList($i)]} {continue}

    set Addr [first [split $XportList([set i]) :]]
    set Port [second [split $XportList([set i]) :]]
    set Active [third [split $XportList([set i]) :]]
    if {$Addr != "" && $Active == ""} {set Active 1}

    if {$Active != 1} {continue}

    puts "Using Digitizer with Xport at $Addr:$Port"
    incr NumDigitizer
    set InUse($i) 1
}


# Use namespaces to isolate your methods
namespace eval urn:AGATADigitizerAccessServer {


    proc AccessServerVerbose {option} {
        global MIDAS_AGATADigitizerAccessServer_Verbose
        global MIDAS_RegisterAccessServer_Verbose
        global MIDAS_SOAPDomainHandler_Verbose

        upvar #0 AGATADigitizerAccessServer_Connected Connected
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
        upvar #0 AGATADigitizerAccessServer_InUse InUse

        set MIDAS_AGATADigitizerAccessServer_Verbose $option
#        set MIDAS_RegisterAccessServer_Verbose $option
#        set MIDAS_SOAPDomainHandler_Verbose $option

        for {set i 0} {$i < $MaxDigitizer} {incr i} {
            set Digitizer $i
            if {$InUse($Digitizer) == 1 && $Connected($Digitizer) == 0} {
                XportAccess_Select $Digitizer
                XportAccess_Verbose $option
            }
        }

        return [rpcvar::rpcvar int 0]
    }

    proc SetTimeOut {Digitizer option} {
        do_SetTimeOut $Digitizer $option
        return [rpcvar::rpcvar int 0]
    }

    proc Init {} {
        set vl [do_Init]
        return [rpcvar::rpcvar string $vl]
    }

    proc Set {Digitizer Fpga Address AddressModifier Value} {
        do_Set $Digitizer $Fpga $Address $AddressModifier 1 $Value
        return [rpcvar::rpcvar int 0]
    }

    proc Write {Digitizer Fpga Address AddressModifier Length Values} {
        do_Set $Digitizer $Fpga $Address $AddressModifier $Length $Values
        return [rpcvar::rpcvar int 0]
    }

    proc Get {Digitizer Fpga Address AddressModifier} {

        switch $AddressModifier {
            A16  {
                set vl [do_Get $Digitizer $Fpga $Address A16]
            }
            default  {
                set vl [do_Read $Digitizer $Fpga $Address $AddressModifier 1]
            }
        }
        return [rpcvar::rpcvar int $vl]
    }

    proc Read {Digitizer Fpga Address AddressModifier Length} {

        switch $AddressModifier {
            A16  {
                if {$Length == 1} {
                    set vl [do_Get $Digitizer $Fpga $Address A16]
                    return [rpcvar::rpcvar int $vl]
                }
            }
            default {}
        }
        set vl [do_Read $Digitizer $Fpga $Address $AddressModifier $Length]
        return [rpcvar::rpcvar string $vl]
    }

    proc do_Init {} {
        upvar #0 AGATADigitizerAccessServer_NumDigitizer NumDigitizer
        upvar #0 AGATADigitizerAccessServer_Connected Connected
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
        upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut
        upvar #0 AGATADigitizerAccessServer_InUse InUse

        for {set i 0} {$i < $MaxDigitizer} {incr i} {set Connected($i) 0; set TimeOut($i) 0}

        set DigitizerList {}
        for {set i 0} {$i < $MaxDigitizer} {incr i} {
            if {$InUse($i) == 0} {continue}
            lappend DigitizerList $i
        }

        return $DigitizerList
    }

    proc do_SetTimeOut {Digitizer option} {
        upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer

        set TimeOut($Digitizer) $option
        XportAccess_Configure -t $option       ;# option = 0 will reset timeout to default
    }

    proc do_Set {args} {
        upvar #0 MIDAS_AGATADigitizerAccessServer_Verbose Verbose
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
        upvar #0 AGATADigitizerAccessServer_NumDigitizer NumDigitizer
        upvar #0 AGATADigitizerAccessServer_InUse InUse
        upvar #0 AGATADigitizerAccessServer_XportList XportList
        upvar #0 AGATADigitizerAccessServer_Connected Connected
        upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut

        set Digitizer [lindex $args 0]
        set Fpga [lindex $args 1]
        set Address [lindex $args 2]
        set AddressModifier [lindex $args 3]
        set Length [lindex $args 4]
        set Values [lreplace $args 0 4]

        if {$Verbose} {puts "called method AGATADigitizerAccessServer::Set with parameters \
               $Digitizer $Fpga $Address $AddressModifier $Length $Values"}
  
#    Check call parameters

       if {$Length != [llength $Values]} then {
           return -code error -errorcode 0x10000 -errorinfo "Incompatible parameters"
       }

#    Digitizer

       if {$Digitizer < 0 || $Digitizer >= $MaxDigitizer || $InUse($Digitizer) == 0} then {
           return -code error -errorcode 0x10001 -errorinfo "Digitizer specified is invalid"
       }

#    FPGA
#      0   core  - adc card 1
#      1   core  - adc card 2
#      2   core  - core adcs
#      3   core  - main board

#      8   segment  - adc card 1
#      9   segment  - adc card 2
#      10  segment  - adc card 3
#      11  segment  - adc card 4
#      12  segment  -  main board

       switch $Fpga {
          0 -
          1 - 
          2 -
          3  {}

          8 - 
          9 - 
          10 -
          11 -
          12 {}

          default {
              return -code error -errorcode 0x10002 -errorinfo "FPGA specified is invalid"
          }
       }

#    Address  -   currently no available checks
#      Note for Main Board Accesses  (Fpga = 3 and 12)  this parameter is the Command

#    AddressModifier    -   all accesses are D16 
#      Note this parameter is ignored for Main Board Accesses  (Fpga = 3 and 12)

        if {$Fpga == 3 || $Fpga == 12} {
           set AddrMod 0
        } else {

           switch $AddressModifier {
               A16  {set AddrMod 0}
               A32  {set AddrMod 1}
               default {
                   return -code error -errorcode 0x10003 -errorinfo "Address Modifier is invalid"
               }
           }
        }

#    now start access

#    check status of TCP connection to Xport interface for the required Digitizer

        if {$Connected($Digitizer) == 0} {

            package require XportAccess

            XportAccess_Select $Digitizer
            XportAccess_Verbose $Verbose

            set XportHost [first [split $XportList($Digitizer) :]]
            set XportPort [second [split $XportList($Digitizer) :]]
            XportAccess_Configure -n $XportHost -p $XportPort 

            XportAccess_Init

            if {$Verbose} {puts "Access to Xport for digitizer $Digitizer at $XportHost:$XportPort initialised"}

            set Connected($Digitizer) 1
        } else {

            XportAccess_Select $Digitizer
        }

      if {$Fpga == 3 || $Fpga == 12} {

#    Access to main board  -  perform write

          if {$Length <= 2} {
              set byte0 [format "0x%x" [expr ($Fpga/8) << 7]]                       ;# routing + simple write
              set byte1 0x0
              set byte2 0x0
              set byte3 0x04       ;# bytes following
              set cmd0  [format "0x%x" [expr (($Fpga/8) << 7) + (($Fpga & 7) << 2)]]  ;# routing + simple write
           } else {
              set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5)]]                     ;# routing + long write
              set byte1 0x0
              set byte2 0x0
              set byte3 [format "0x%x" [expr 2 + $Length]]    ;# bytes following
              set cmd0  [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5) + (($Fpga & 7) << 2)]]  ;# routing + long write
          }

          set cmd1 [format "0x%x" [expr ($Address & 255)]]

          if {$Length <= 2} {
              if {$Length == 1} {
                  set Value [first $Values]
                  set data0 [format "0x%x" [expr ($Value >> 8) & 255]]
                  set data1 [format "0x%x" [expr ($Value & 255)]]
              } else {
                  set data0 [format "0x%x" [first $Values]]
                  set data1 [format "0x%x" [second $Values]]
              }
              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1"}
              set z [catch {XportAccess_Set        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1} m]
          } else {
              set data ""
              foreach Value $Values {append data [format " 0x%x" [expr ($Value & 255)]]}
              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 ...."}
              set z [catch {eval XportAccess_Write   $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data} m]
          }

      } else {

       if {$AddrMod == 0} {

#     A16 access

          set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5)]]   ;# routing + long write
          set byte1 0x0
          set byte2 0x0
          set byte3 [format "0x%x" [expr 4 + ($Length * 2)]]    ;# bytes following
          set cmd0  [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5) + (($Fpga & 7) << 2)]]  ;# routing + long write
          set cmd1  [format "0x%x" 0]      ;#  A16 write command 
          set data0 [format "0x%x" [expr ($Address >> 8) & 255]]
          set data1 [format "0x%x" [expr ($Address & 255)]]

          if {$Length == 1} {
              set Value [first $Values]
              set data2 [format "0x%x" [expr ($Value >> 8) & 255]]
              set data3 [format "0x%x" [expr ($Value & 255)]]

              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3"}
              set z [catch {XportAccess_Write        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3} m]
          } else {
              set data ""
              foreach Value $Values {
                  append data [format " 0x%x" [expr ($Value >> 8) & 255]]
                  append data [format " 0x%x" [expr ($Value & 255)]]
              }
              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 ...."}
              set z [catch {eval XportAccess_Write   $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data} m]
          }

       } else {

#     A32 access

          set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5)]]   ;# routing + long write
          set byte1 0x0
          set byte2 0x0
          set byte3 [format "0x%x" [expr 6 + ($Length * 2)]]    ;# bytes following
          set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5) + (($Fpga & 7) << 2)]]  ;# routing + long write
          set cmd1 [format "0x%x" 1]      ;#  A32 write command 
          set data0 [format "0x%x" [expr ($Address >> 24) & 255]]
          set data1 [format "0x%x" [expr ($Address >> 16) & 255]]
          set data2 [format "0x%x" [expr ($Address >> 8) & 255]]
          set data3 [format "0x%x" [expr ($Address & 255)]]

          if {$Length == 1} {
              set Value [first $Values]
              set data4 [format "0x%x" [expr ($Value >> 8) & 255]]
              set data5 [format "0x%x" [expr ($Value & 255)]]

              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 $data4 $data5"}
              set z [catch {XportAccess_Write        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 $data4 $data5} m]
          } else {
              set data ""
              foreach Value $Values {
                  append data [format " 0x%x" [expr ($Value >> 8) & 255]]
                  append data [format " 0x%x" [expr ($Value & 255)]]
              }
              if {$Verbose} {puts "sending Write command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 ...."}
              set z [catch {eval XportAccess_Write   $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 $data} m]
          }

       }

       }

       if {$z == 0} then {
           if {$Verbose} {puts "received good response:\n $m"}
           set response [last [split $m \n]]
           if {[llength $response] == 4} then {
               if {$Verbose} {puts "received Successful Write ACK\n$response"}
               return 0
           }
           if {[llength $response] == 6} then {
               if {$Verbose} {puts "received Failed Write ACK\n$response"}
               return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
           }
           if {$Verbose} {puts "received unexpected reply: $response"}
           set Connected($Digitizer) 0
           return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
       } else {
           if {$Verbose} {puts "received error response: rc=$z\n  $m"}
           set Connected($Digitizer) 0
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
       }

    }

    proc do_Read {Digitizer Fpga Address AddressModifier Length} {
        upvar #0 MIDAS_AGATADigitizerAccessServer_Verbose Verbose
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
        upvar #0 AGATADigitizerAccessServer_NumDigitizer NumDigitizer
        upvar #0 AGATADigitizerAccessServer_InUse InUse
        upvar #0 AGATADigitizerAccessServer_XportList XportList
        upvar #0 AGATADigitizerAccessServer_Connected Connected
        upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut

        if {$Verbose} {puts "called method AGATADigitizerAccessServer::Read with parameters \
$Digitizer $Fpga $Address $AddressModifier $Length"}

#    Check call parameters 

#    Digitizer

       if {$Digitizer < 0 || $Digitizer >= $MaxDigitizer || $InUse($Digitizer) == 0} then {
           return -code error -errorcode 0x10001 -errorinfo "Digitizer specified is invalid"
       }

#    FPGA

       switch $Fpga {
          0 -
          1 - 
          2 -
          3  {}

          8 - 
          9 - 
          10 -
          11 -
          12 {}

          default {
              return -code error -errorcode 0x10002 -errorinfo "FPGA specified is invalid"
          }
       }

#    Address  -   currently no available checks
#      Note for Main Board Accesses  (Fpga = 3 and 12)  this parameter is the Command

#    AddressModifier    -   all accesses are D16 
#      Note this parameter is ignored for Main Board Accesses  (Fpga = 3 and 12)

        if {$Fpga == 3 || $Fpga == 12} {
           set AddrMod 0
        } else {

           switch $AddressModifier {
               A16  {set AddrMod 0}
               A32  {set AddrMod 1}
               default {
                   return -code error -errorcode 0x10003 -errorinfo "Address Modifier is invalid"
               }
           }
        }

#    now start access

#    check status of TCP connection to Xport interface for the required Digitizer

        if {$Connected($Digitizer) == 0} {

            package require XportAccess

            XportAccess_Select $Digitizer
            XportAccess_Verbose $Verbose

            set XportHost [first [split $XportList($Digitizer) :]]
            set XportPort [second [split $XportList($Digitizer) :]]
            XportAccess_Configure -n $XportHost -p $XportPort  

            XportAccess_Init

            if {$Verbose} {puts "Access to Xport for digitizer $Digitizer at $XportHost:$XportPort initialised"}

            set Connected($Digitizer) 1
        } else {

            XportAccess_Select $Digitizer
        }

      if {$Fpga == 3 || $Fpga == 12} {

#    Access to main board  -  perform basic simple read

          set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5)]]   ;# routing + read
          set byte1 0x0
          set byte2 0x0
          set byte3 0x4    ;# bytes following
          set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5) + (($Fpga & 7) << 2)]]
          set cmd1 [format "0x%x" $Address]
          set data0 0x0
          set data1 0x0

          if {$Verbose} {puts "sending Get command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1"}

          set z [catch {XportAccess_Get        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1} m]
          if {$z == 0} then {
             if {$Verbose} {puts "received good response:\n $m"}
             set response [last [split $m \n]]
             if {[llength $response] > 6} then {
                if {$Verbose} {puts "received Successful Read ACK\n$response"}
                set Values ""
                for {set i 6} {$i < [llength $response]} {incr i} {
                    set byte [expr [lindex $response $i] & 255]
                    append Values [format "0x%x " $byte]
                }
             } else {
                 if {[llength $response] == 6} then {
                    if {$Verbose} {puts "received Failed Read ACK\n$response"}
                    return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
                 }
                 if {$Verbose} {puts "received unexpected reply: $response"}
                 set Connected($Digitizer) 0
                 return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
             }
          } else {
             if {$Verbose} {puts "received error response: rc=$z\n  $m"}
             set Connected($Digitizer) 0
             return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
          }

      } else {

#    Read access

#    first perform setup

       set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5)]]   ;# routing + long write
       set byte1 0x0
       set byte2 0x0
       if {$AddrMod == 0} {
           set byte3 0x6    ;# bytes following
       } else {
           set byte3 0x8    ;# bytes following
       }
       set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (1 << 5) + (($Fpga & 7) << 2)]]
       if {$AddrMod == 0} {
            set cmd1 [format "0x%x" 16]      ;#  A16 read setup command
            set data0 [format "0x%x" [expr ($Address >> 8) & 255]]
            set data1 [format "0x%x" [expr ($Address & 255)]]
            set data2 [format "0x%x" [expr ($Length >> 8) & 255]]
            set data3 [format "0x%x" [expr ($Length & 255)]]
            if {$Verbose} {puts "sending Read Setup command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3"}
            set z [catch {XportAccess_Write        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3} m]
       } else {            
            set cmd1 [format "0x%x" 17]      ;#  A32 read setup command
            set data0 [format "0x%x" [expr ($Address >> 24) & 255]]
            set data1 [format "0x%x" [expr ($Address >> 16) & 255]]
            set data2 [format "0x%x" [expr ($Address >> 8) & 255]]
            set data3 [format "0x%x" [expr ($Address & 255)]]
            set data4 [format "0x%x" [expr ($Length >> 8) & 255]]
            set data5 [format "0x%x" [expr ($Length & 255)]]
            if {$Verbose} {puts "sending Read Setup command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 $data4 $data5"}
            set z [catch {XportAccess_Write        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1 $data2 $data3 $data4 $data5} m]
       }
       if {$z == 0} then {
           if {$Verbose} {puts "received good response:\n $m"}
           set response [last [split $m \n]]
           if {[llength $response] == 4} then {
               if {$Verbose} {puts "received Successful Write ACK\n$response"}
           } else {
              if {[llength $response] == 6} then {
                  if {$Verbose} {puts "received Failed Write ACK\n$response"}
                  return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
              }
              if {$Verbose} {puts "received unexpected reply: $response"}
              set Connected($Digitizer) 0
              return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
           }
       } else {
           if {$Verbose} {puts "received error response: rc=$z\n  $m"}
           set Connected($Digitizer) 0
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
       }

#    now perform data access

       set Values ""

        for {set i 0} {$i < $Length} {incr i} {

           set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5)]]   ;# routing + read
           set byte1 0x0
           set byte2 0x0
           set byte3 0x4    ;# bytes following
           set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5) + (($Fpga & 7) << 2)]]
           if {$AddrMod == 0} {
               set cmd1 [format "0x%x" 0]      ;#  A16 read command
           } else {
               set cmd1 [format "0x%x" 1]      ;#  A32 read command
           }
           set data0 0x0
           set data1 0x0

           if {$Verbose} {puts "sending Block Read command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1"}
           set z [catch {XportAccess_Get        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1} m]
           if {$z == 0} then {
               if {$Verbose} {puts "received good response:\n $m"}
               set response [last [split $m \n]]
               if {[llength $response] == 8} then {
                   if {$Verbose} {puts "received Successful Read ACK\n$response"}
                   set Value [expr (([lindex $response 6] & 255) << 8) + ([lindex $response 7] & 255)]
                   append Values "$Value "
               } else {
                  if {[llength $response] == 6} then {
                      if {$Verbose} {puts "received Failed Read ACK\n$response"}
                      return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
                  }
                  if {$Verbose} {puts "received unexpected reply: $response"}
                  set Connected($Digitizer) 0
                  return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
               }
           } else {
               if {$Verbose} {puts "received error response: rc=$z\n  $m"}
               set Connected($Digitizer) 0
               return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
           }
         }
       }
       return $Values
    }

    proc do_Get {Digitizer Fpga Address AddressModifier} {
        upvar #0 MIDAS_AGATADigitizerAccessServer_Verbose Verbose
        upvar #0 AGATADigitizerAccessServer_MaxDigitizer MaxDigitizer
        upvar #0 AGATADigitizerAccessServer_NumDigitizer NumDigitizer
        upvar #0 AGATADigitizerAccessServer_InUse InUse
        upvar #0 AGATADigitizerAccessServer_XportList XportList
        upvar #0 AGATADigitizerAccessServer_Connected Connected
        upvar #0 AGATADigitizerAccessServer_TimeOut TimeOut

        if {$Verbose} {puts "called method AGATADigitizerAccessServer::Get with parameters \
               $Digitizer $Fpga $Address $AddressModifier"}

#    Check call parameters 

#    Digitizer

       if {$Digitizer < 0 || $Digitizer >= $MaxDigitizer || $InUse($Digitizer) == 0} then {
           return -code error -errorcode 0x10001 -errorinfo "Digitizer specified is invalid"
       }

#    FPGA

       switch $Fpga {
          0 -
          1 - 
          2 -
          3  {}

          8 - 
          9 - 
          10 -
          11 -
          12 {}

          default {
              return -code error -errorcode 0x10002 -errorinfo "FPGA specified is invalid"
          }
       }

#    Address  -   currently no available checks
#      Note for Main Board Accesses  (Fpga = 3 and 12)  this parameter is the Command

#    AddressModifier    -   all accesses are D16    (only A16 supported by simple access)
#      Note this parameter is ignored for Main Board Accesses  (Fpga = 3 and 12)

        if {$Fpga == 3 || $Fpga == 12} {
           set AddrMod 0
        } else {

           switch $AddressModifier {
               A16  {set AddrMod 0}
               default {
                   return -code error -errorcode 0x10003 -errorinfo "Address Modifier is invalid"
               }
           }
        }

#    now start access

#    check status of TCP connection to Xport interface for the required Digitizer

        if {$Connected($Digitizer) == 0} {

            package require XportAccess

            XportAccess_Select $Digitizer
            XportAccess_Verbose $Verbose

            set XportHost [first [split $XportList($Digitizer) :]]
            set XportPort [second [split $XportList($Digitizer) :]]
            XportAccess_Configure -n $XportHost -p $XportPort  

            XportAccess_Init

            if {$Verbose} {puts "Access to Xport for digitizer $Digitizer at $XportHost:$XportPort initialised"}

            set Connected($Digitizer) 1
        } else {

            XportAccess_Select $Digitizer
        }

       if {$Fpga == 3 || $Fpga == 12} {

#    Access to main board  -  perform basic simple read

          set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5)]]   ;# routing + read
          set byte1 0x0
          set byte2 0x0
          set byte3 0x4    ;# bytes following
          set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5) + (($Fpga & 7) << 2)]]
          set cmd1 [format "0x%x" $Address]
          set data0 0x0
          set data1 0x0

          if {$Verbose} {puts "sending Get command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1"}

          set z [catch {XportAccess_Get        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1} m]
          if {$z == 0} then {
             if {$Verbose} {puts "received good response:\n $m"}
             set response [last [split $m \n]]
             if {[llength $response] == 8} then {
                if {$Verbose} {puts "received Successful Read ACK\n$response"}
                set Value [expr (([lindex $response 6] & 255) << 8) + ([lindex $response 7] & 255)]
                return $Value
             }
             if {[llength $response] == 6} then {
                if {$Verbose} {puts "received Failed Read ACK\n$response"}
                return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
             }
             if {$Verbose} {puts "received unexpected reply: $response"}
             set Connected($Digitizer) 0
             return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
          } else {
             if {$Verbose} {puts "received error response: rc=$z\n  $m"}
             set Connected($Digitizer) 0
             return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
          }

       } else {

#    Get access (single A16 D16 read)

          set byte0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5)]]   ;# routing + read
          set byte1 0x0
          set byte2 0x0
          set byte3 0x4    ;# bytes following
          set cmd0 [format "0x%x" [expr (($Fpga/8) << 7) + (2 << 5) + (($Fpga & 7) << 2)]]
          set cmd1 [format "0x%x" 16]      ;#  A16 read command 
          set data0 [format "0x%x" [expr ($Address >> 8)]]
          set data1 [format "0x%x" [expr ($Address & 255)]]

          if {$Verbose} {puts "sending Get command to Xport : $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1"}

          set z [catch {XportAccess_Get        $byte0 $byte1 $byte2 $byte3 $cmd0 $cmd1 $data0 $data1} m]
          if {$z == 0} then {
             if {$Verbose} {puts "received good response:\n $m"}
             set response [last [split $m \n]]
             if {[llength $response] == 8} then {
                if {$Verbose} {puts "received Successful Read ACK\n$response"}
                set Value [expr (([lindex $response 6] & 255) << 8) + ([lindex $response 7] & 255)]
                return $Value
             }
             if {[llength $response] == 6} then {
                if {$Verbose} {puts "received Failed Read ACK\n$response"}
                return -code error -errorcode 0x10006 -errorinfo "received Failed Read ACK\n$response"
             }
             if {$Verbose} {puts "received unexpected reply: $response"}
             set Connected($Digitizer) 0
             return -code error -errorcode 0x10004 -errorinfo "received unexpected reply: $response"
          } else {
             if {$Verbose} {puts "received error response: rc=$z\n  $m"}
             set Connected($Digitizer) 0
             return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n  $m"
          }

       }

    }



    # We have to publish the public methods...

     SOAP::export AccessServerVerbose
     SOAP::export SetTimeOut
     SOAP::export Init
     SOAP::export Set
     SOAP::export Get
     SOAP::export Write
     SOAP::export Read

     
}   ;#  end of    namespace eval urn:AGATADigitizerAccessServer

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /AGATADigitizerAccessServer -namespace urn:AGATADigitizerAccessServer -uri urn:AGATADigitizerAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/AGATADigitizerAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}


set msg "Web Service Installed\n"
for {set i 0} {$i < $MaxDigitizer} {incr i} {
    if {![info exists XportList($i)]} {continue}
    if {$InUse($i) == 0} {continue}

    set Addr [first [split $XportList([set i]) :]]
    set Port [second [split $XportList([set i]) :]]

    append msg  "Using Digitizer with Xport at $Addr:$Port\n"
}
DefineMessage -c digitizer -l INFORMATION  "$msg"
